# BEGIN_COPYRIGHT
#
# Copyright (C) 2020-2023 Paradigm4 Inc.
# All Rights Reserved.
#
# scidbbridge is a plugin for SciDB, an Open Source Array DBMS
# maintained by Paradigm4. See http://www.paradigm4.com/
#
# scidbbridge is free software: you can redistribute it and/or modify
# it under the terms of the AFFERO GNU General Public License as
# published by the Free Software Foundation.
#
# scidbbridge is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY
# KIND, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
# NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See the
# AFFERO GNU General Public License for the complete license terms.
#
# You should have received a copy of the AFFERO GNU General Public
# License along with scidbbridge. If not, see
# <http://www.gnu.org/licenses/agpl-3.0.html>
#
# END_COPYRIGHT

import itertools
import os.path
import pandas
import pyarrow
import pytest
import requests

from conftest import *


@pytest.mark.parametrize('format, compression',
                         itertools.product(('arrow', 'parquet'),
                                           ('none', 'gzip', 'lz4')))
def test_one_chunk(scidb_con, format, compression):
    url = '{}/xsave/one_chunk/{}/{}'.format(test_url, format, compression)
    schema = '<v:int64> [i=0:9]'

    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}',
  format:'{}',
  compression:'{}')""".format(schema, url, format, compression))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {
        **base_metadata,
        **{'schema': '{}'.format(
            schema.replace(']', ':0:1000000]')),
           'format': format,
           'compression': None if compression == 'none' else compression}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data={'i': range(1)}))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0).to_pandas(),
        pandas.DataFrame(data={'v': range(10), 'i': range(10)}))


@pytest.mark.parametrize('format, compression',
                         itertools.product(('arrow', 'parquet'),
                                           ('none', 'gzip', 'lz4')))
def test_multi_chunk(scidb_con, format, compression):
    url = '{}/xsave/multi_chunk/{}/{}'.format(test_url, format, compression)
    schema = '<v:int64> [i=0:19:0:5]'

    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}',
  format:'{}',
  compression:'{}')""".format(schema, url, format, compression))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {
        **base_metadata,
        **{'schema': '{}'.format(schema),
           'format': format,
           'compression': None if compression == 'none' else compression}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data={'i': range(0, 20, 5)}))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0).to_pandas(),
        pandas.DataFrame(data={'v': range(5), 'i': range(5)}))
    pandas.testing.assert_frame_equal(
        array.get_chunk(10).to_pandas(),
        pandas.DataFrame(data={'v': range(10, 15), 'i': range(10, 15)}))


@pytest.mark.parametrize('format, compression',
                         itertools.product(('arrow', 'parquet'),
                                           ('none', 'gzip', 'lz4')))
def test_multi_dim(scidb_con, format, compression):
    url = '{}/xsave/multi_dim/{}/{}'.format(test_url, format, compression)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}',
  format:'{}',
  compression:'{}')""".format(schema, url, format, compression))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {
        **base_metadata,
        **{'schema': '{}'.format(schema),
           'format': format,
           'compression': None if compression == 'none' else compression}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 9, 5)
                               for j in range(10, 20, 5)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 10).to_pandas(),
        pandas.DataFrame(data=((i, i, j)
                               for i in range(5)
                               for j in range(10, 15)),
                         columns=('v', 'i', 'j')))


@pytest.mark.parametrize('format, compression',
                         itertools.product(('arrow', 'parquet'),
                                           ('none', 'gzip', 'lz4')))
def test_multi_atts(scidb_con, format, compression):
    url = '{}/xsave/multi_attr/{}/{}'.format(test_url, format, compression)
    schema = '<v:int64,w:double> [i=0:9:0:5; j=10:19:0:5]'

    scidb_con.iquery("""
xsave(
  apply(
    build({}, i),
    w, double(v * v)),
  '{}',
  format:'{}',
  compression:'{}')""".format(schema.replace(',w:double', ''),
                              url,
                              format,
                              compression))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {
        **base_metadata,
        **{'schema': '{}'.format(schema),
           'format': format,
           'compression': None if compression == 'none' else compression}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 9, 5)
                               for j in range(10, 20, 5)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 10).to_pandas(),
        pandas.DataFrame(data=((i, float(i * i), i, j)
                               for i in range(5)
                               for j in range(10, 15)),
                         columns=('v', 'w', 'i', 'j')))


def test_filter(scidb_con):
    url = '{}/xsave/filter'.format(test_url)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    scidb_con.iquery("""
xsave(
  filter(
    build({}, i),
    (i < 3 or i > 5) and j > 15),
  '{}')""".format(schema, url))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, 15) for i in range(0, 9, 5)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 15).to_pandas(),
        pandas.DataFrame(data=((i, i, j)
                               for i in range(3)
                               for j in range(16, 20)),
                         columns=('v', 'i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(5, 15).to_pandas(),
        pandas.DataFrame(data=((i, i, j)
                               for i in range(6, 10)
                               for j in range(16, 20)),
                         columns=('v', 'i', 'j')))


def test_empty_chunks(scidb_con):
    url = '{}/xsave/empty_chunks'.format(test_url)
    schema = '<v:int64> [i=0:19:0:5; j=10:49:0:10]'

    scidb_con.iquery("""
xsave(
  filter(
    build({}, i + j),
    (i >= 5 and i < 10 or i >= 15) and (j < 20 or j >= 30 and j < 40)),
  '{}')""".format(schema, url))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in (5, 15)
                               for j in (10, 30)),
                         columns=('i', 'j')))

    for i_st in (5, 15):
        for j_st in (10, 30):
            pandas.testing.assert_frame_equal(
                array.get_chunk(i_st, j_st).to_pandas(),
                pandas.DataFrame(data=((i + j, i, j)
                                       for i in range(i_st, i_st + 5)
                                       for j in range(j_st, j_st + 10)),
                                 columns=('v', 'i', 'j')))


def test_one_index(scidb_con):
    prefix = 'one_index'
    url = '{}/xsave/{}'.format(test_url, prefix)
    schema = '<v:int64> [i=0:99:0:5; j=0:99:0:5]'

    scidb_con.iquery("""
xsave(
  build({}, i + j),
  '{}')""".format(schema, url))

    array = scidbbridge.Array(url)

    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 100, 5)
                               for j in range(0, 100, 5)),
                         columns=('i', 'j')))

    parts = urllib.parse.urlparse(test_url)
    if parts.scheme == 's3':
        index_prefix = '{}/xsave/{}/index/'.format(parts.path[1:], prefix)
        result = s3_con.list_objects_v2(Bucket=parts.netloc,
                                        Prefix=index_prefix)
        keys = [split['Key'] for split in result['Contents']]
    elif parts.scheme == 'file':
        index_prefix = '{}/xsave/{}/index/'.format(parts.path, prefix)
        keys = []
        for fn in os.listdir(index_prefix):
            key = index_prefix + fn
            if os.path.isfile(key):
                keys.append(key)
        keys.sort()

    assert keys == ['{}{}'.format(index_prefix, 0)]


@pytest.mark.parametrize('index_split', (100, 200, 400, 800))
def test_multi_index(scidb_con, index_split):
    prefix = 'multi_index_{}'.format(index_split)
    url = '{}/xsave/{}'.format(test_url, prefix)
    schema = '<v:int64> [i=0:99:0:5; j=0:99:0:5]'

    scidb_con.iquery("""
xsave(
  build({}, i + j),
  '{}',
  index_split:{})""".format(schema, url, index_split))

    array = scidbbridge.Array(url)

    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 100, 5)
                               for j in range(0, 100, 5)),
                         columns=('i', 'j')))

    parts = urllib.parse.urlparse(test_url)
    if parts.scheme == 's3':
        index_prefix = '{}/xsave/{}/index/'.format(parts.path[1:], prefix)
        result = s3_con.list_objects_v2(Bucket=parts.netloc,
                                        Prefix=index_prefix)
        keys = [split['Key'] for split in result['Contents']]
    elif parts.scheme == 'file':
        index_prefix = '{}/xsave/{}/index/'.format(parts.path, prefix)
        keys = []
        for fn in os.listdir(index_prefix):
            key = index_prefix + fn
            if os.path.isfile(key):
                keys.append(key)
        keys.sort()

    assert keys == ['{}{}'.format(index_prefix, i)
                    for i in range(800 // index_split)]


@pytest.mark.parametrize('format', ('default', 'arrow', 'parquet'))
def test_format(scidb_con, format):
    prefix = 'format_{}'.format(format)
    url = '{}/xsave/{}'.format(test_url, prefix)
    schema = '<v:int64> [i=0:19:0:5; j=10:49:0:10]'

    scidb_con.iquery("""
xsave(
  build({}, i + j),
  '{}'{})""".format(schema,
                    url,
                    '' if format == 'default'
                    else ", format:'{}'".format(format)))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema),
                                 'format': '{}'.format(
                                     'arrow' if format == 'default'
                                     else format)}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 20, 5)
                               for j in range(10, 50, 10)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 10).to_pandas(),
        pandas.DataFrame(data=((i + j, i, j)
                               for i in range(0, 5)
                               for j in range(10, 20)),
                         columns=('v', 'i', 'j')))


@pytest.mark.parametrize('compression, sz_min, sz_max',
                         zip(('default', 'none', 'gzip', 'lz4'),
                             (0, 1000, 0, 0),
                             (500, 9999, 500, 500)))
def test_compression(scidb_con, compression, sz_min, sz_max):
    prefix = 'compression_{}'.format(compression)
    url = '{}/xsave/{}'.format(test_url, prefix)
    schema = '<v:int64> [i=0:19:0:5; j=10:49:0:10]'

    scidb_con.iquery("""
xsave(
  build({}, i + j),
  '{}'{})""".format(schema,
                    url,
                    '' if compression == 'default'
                    else ", compression:'{}'".format(compression)))

    array = scidbbridge.Array(url)

    m_compression = (None if compression == 'none'
                     else 'lz4' if compression == 'default'
                     else compression)
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema),
                                 'compression': m_compression}}
    for i_st in (5, 15):
        for j_st in (10, 30):
            pandas.testing.assert_frame_equal(
                array.get_chunk(i_st, j_st).to_pandas(),
                pandas.DataFrame(data=((i + j, i, j)
                                       for i in range(i_st, i_st + 5)
                                       for j in range(j_st, j_st + 10)),
                                 columns=('v', 'i', 'j')))

    parts = urllib.parse.urlparse(test_url)
    if parts.scheme == 's3':
        sz = s3_con.head_object(Bucket=parts.netloc,
                                Key='{}/xsave/{}/chunks/c_0_0'.format(
                                    parts.path[1:], prefix))['ContentLength']
    elif parts.scheme == 'file':
        sz = os.path.getsize('{}/xsave/{}/chunks/c_0_0'.format(parts.path,
                                                               prefix))

    assert sz > sz_min
    assert sz < sz_max


@pytest.mark.parametrize('format, compression',
                         itertools.product(('default', 'arrow', 'parquet'),
                                           ('default', 'none', 'gzip', 'lz4')))
def test_format_compression(scidb_con, format, compression):
    prefix = 'format_compression/{}/{}'.format(format, compression)
    url = '{}/xsave/{}'.format(test_url, prefix)
    schema = '<v:int64> [i=0:19:0:5; j=10:49:0:10]'

    scidb_con.iquery("""
xsave(
  build({}, i + j),
  '{}'{}{})""".format(schema,
                      url,
                      '' if format == 'default'
                      else ", format:'{}'".format(format),
                      '' if compression == 'default'
                      else ", compression:'{}'".format(compression)))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    m_compression = (None if compression == 'none'
                     else 'lz4' if compression == 'default'
                     else compression)
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema),
                                 'format': '{}'.format(
                                     'arrow' if format == 'default'
                                     else format),
                                 'compression': m_compression}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 20, 5)
                               for j in range(10, 50, 10)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 10).to_pandas(),
        pandas.DataFrame(data=((i + j, i, j)
                               for i in range(0, 5)
                               for j in range(10, 20)),
                         columns=('v', 'i', 'j')))


@pytest.mark.parametrize('url', (None,
                                 '',
                                 'foo',
                                 'foo://',
                                 'foo://bar/taz',
                                 's3',
                                 's3:',
                                 's3:/',
                                 's3://',
                                 's3:\\',
                                 's3:\\\\',
                                 'file',
                                 'file:',
                                 'file:/',
                                 'file://',
                                 'file:\\',
                                 ))
def test_bad_url(scidb_con, url):
    schema = '<v:int64> [i=0:9]'

    que = 'xsave(build({}, i){})'.format(
        schema, '' if url is None else ", '{}'".format(url))

    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(que)


def test_update_error(scidb_con):
    name = 'update_error_{}'.format(test_id)
    url = '{}/xsave/update_error'.format(test_url)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # No "update:true"
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # "update:false"
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build({}, i),
  '{}', update:false)""".format(schema, url))

    # "update:'foo'"
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build({}, i),
  '{}', update:'foo')""".format(schema, url))

    # Mismatched lower dimension
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=-1:9:0:5; j=10:19:0:5], i),
  '{}', update:true)""".format(url))

    # Mismatched higher dimension
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:0:5], i),
  '{}', update:true)""".format(url))

    # Mismatched chunk size
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:10; j=10:20:0:5], i),
  '{}', update:true)""".format(url))

    # Mismatched chunk overlap
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:5:5], i),
  '{}', update:true)""".format(url))

    # Set format
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:0:5], i),
  '{}', update:true, format:'arrow')""".format(url))

    # Set format
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:0:5], i),
  '{}', update:true, format:'parquet')""".format(url))

    # Set compression
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:0:5], i),
  '{}', update:true, compression:'gzip')""".format(url))

    # Set compression
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:0:5], i),
  '{}', update:true, compression:'lz4')""".format(url))

    # Set index_split
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:0:5], i),
  '{}', update:true, index_split:10000)""".format(url))

    # Set namespace
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build(<v:int64> [i=0:9:0:5; j=10:20:0:5], i),
  '{}', update:true, namespace:public)""".format(url))

    # Existing array w/ different attribute type
    scidb_con.iquery("""
store(
  build(<v:double> [i=0:9:0:5; j=10:19:0:5], i),
  {})""".format(name))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave({}, '{}', update:true)".format(name, url))
    scidb_con.remove(name)

    # Existing array w/ different attribute name
    scidb_con.iquery("""
store(
  build(<w:int64> [i=0:9:0:5; j=10:19:0:5], i),
  {})""".format(name))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave({}, '{}', update:true)".format(name, url))
    scidb_con.remove(name)

    # Existing array w/ different number of attributes
    scidb_con.iquery("""
store(
  apply(
    build(<v:int64> [i=0:9:0:5; j=10:19:0:5], i),
    w, double(j)),
  {})""".format(name))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave({}, '{}', update:true)".format(name, url))
    scidb_con.remove(name)

    # Existing array w/ different number of dimensions
    scidb_con.iquery("""
store(
  build(<v:int64> [i=0:9:0:5], i),
  {})""".format(name))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave({}, '{}', update:true)".format(name, url))
    scidb_con.remove(name)

    # Existing array w/ different dimension bounds
    scidb_con.iquery("""
store(
  build(<v:int64> [i=0:19:0:5; j=10:19:0:5], i),
  {})""".format(name))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave({}, '{}', update:true)".format(name, url))
    scidb_con.remove(name)

    # Existing array w/ different dimension chunk size
    scidb_con.iquery("""
store(
  build(<v:int64> [i=0:9:0:5; j=10:19:0:10], i),
  {})""".format(name))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave({}, '{}', update:true)".format(name, url))
    scidb_con.remove(name)


def test_update_array(scidb_con):
    name = 'update_array_{}'.format(test_id)
    url = '{}/xsave/update_array'.format(test_url)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    scidb_con.iquery("""
store(
  build({}, i + 1),
  {})""".format(schema, name))

    scidb_con.iquery("xsave({}, '{}', update:true)".format(name, url))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 9, 5)
                               for j in range(10, 20, 5)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 10).to_pandas(),
        pandas.DataFrame(data=((i + 1, i, j)
                               for i in range(5)
                               for j in range(10, 15)),
                         columns=('v', 'i', 'j')))

    # Cleanup
    scidb_con.remove(name)


def test_update_all(scidb_con):
    url = '{}/xsave/update_all'.format(test_url)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    scidb_con.iquery("""
xsave(
  build({}, i + 1),
  '{}', update:true)""".format(schema, url))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 9, 5)
                               for j in range(10, 20, 5)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 10).to_pandas(),
        pandas.DataFrame(data=((i + 1, i, j)
                               for i in range(5)
                               for j in range(10, 15)),
                         columns=('v', 'i', 'j')))


def test_update_filter(scidb_con):
    url = '{}/xsave/update_filter'.format(test_url)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    scidb_con.iquery("""
xsave(
  filter(
    build({}, i),
    (i < 5 or j >= 15)and i % 3 = 0),
  '{}')""".format(schema, url))

    scidb_con.iquery("""
xsave(
  filter(
    build({}, i),
    i % 2 = 0),
  '{}', update:true)""".format(schema, url))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(0, 9, 5)
                               for j in range(10, 20, 5)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 10).to_pandas(),
        pandas.DataFrame(data=((i, i, j)
                               for i in range(5)
                               for j in range(10, 15)
                               if ((i < 5 or j >= 15) and i % 3 == 0 or
                                   i % 2 == 0)),
                         columns=('v', 'i', 'j')))


def test_update_filter_index(scidb_con):
    url = '{}/xsave/update_filter_index'.format(test_url)
    schema = '<v:int64> [i=0:2:0:1; j=0:2:0:1]'

    scidb_con.iquery("""
xsave(
  filter(
    build({}, i),
    i > 1),
  '{}')""".format(schema, url))

    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}', update:true)""".format(schema, url))

    array = scidbbridge.Array(url)

    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}
    pandas.testing.assert_frame_equal(
        array.read_index(),
        pandas.DataFrame(data=((i, j)
                               for i in range(3)
                               for j in range(3)),
                         columns=('i', 'j')))
    pandas.testing.assert_frame_equal(
        array.get_chunk(0, 0).to_pandas(),
        pandas.DataFrame(data=((i, i, j)
                               for i in range(1)
                               for j in range(1)),
                         columns=('v', 'i', 'j')))


@pytest.mark.parametrize('ty, value',
                         itertools.chain(
                             (('{}int{}'.format(u, s), 'i')
                              for u in ('', 'u')
                              for s in (8, 16, 32, 64)),

                             (('bool', 'i % 2'),
                              ('float', 'i + .1'),
                              ('double', 'i + .01'),
                              ('char', 'char(i)'),
                              ('string', "'foo' + string(i)"),
                              ('datetime', 'datetime(i)'))))
def test_types(scidb_con, ty, value):
    prefix = 'types_{}'.format(ty)
    url = '{}/xsave/{}'.format(test_url, prefix)
    schema = '<v:{}> [i=0:19:0:5; j=10:19:0:5]'.format(ty)

    scidb_con.iquery("""
xsave(
  build({}, iif(i % 3 = 0, null, {})),
  '{}')""".format(schema,
                  value,
                  url))

    array = scidbbridge.Array(url)
    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}


@pytest.mark.parametrize('ty, value',
                         itertools.chain(
                             (('{}int{}'.format(u, s), 'i')
                              for u in ('', 'u')
                              for s in (8, 16, 32, 64)),

                             (('bool', 'i % 2'),
                              ('float', 'i + .1'),
                              ('double', 'i + .01'),
                              ('char', 'char(i)'),
                              ('string', "'foo' + string(i)"),
                              ('datetime', 'datetime(i)'))))
def test_update_types(scidb_con, ty, value):
    prefix = 'update_types_{}'.format(ty)
    url = '{}/xsave/{}'.format(test_url, prefix)
    schema = '<v:{}> [i=0:19:0:5; j=10:19:0:5]'.format(ty)

    scidb_con.iquery("""
xsave(
  filter(
    build({}, iif(i % 3 = 0, null, {})),
    i < 10 and i >= 10),
  '{}')""".format(schema,
                  value,
                  url))

    array = scidbbridge.Array(url)
    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}

    scidb_con.iquery("""
xsave(
  filter(
    build({}, iif(i % 2 = 0, null, {})),
    i >= 5 and i < 15),
  '{}', update:true)""".format(schema,
                               value,
                               url))

    array = scidbbridge.Array(url)
    assert array.__str__() == url
    assert array.metadata == {**base_metadata,
                              **{'schema': '{}'.format(schema)}}


@pytest.mark.skipif(not security_enable,
                    reason="Requires SciDB Enterprise Edition")
def test_permissions(scidb_con):
    # Setup
    scidb_con.create_namespace(test_ns)
    scidb_con.create_user(test_user, test_password_hash)

    con_args = {'scidb_url': scidb_url,
                'scidb_auth': (test_user, test_password),
                'verify': False}
    scidb_con2 = scidbpy.connect(**con_args)

    url = '{}/xsave/permissions'.format(test_url)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    # FS Init
    parts = urllib.parse.urlparse(test_url)
    if parts.scheme == 'file':
        fs_path = os.path.join(parts.path, 'xsave')
        if not os.path.exists(fs_path):
            os.makedirs(fs_path)

    # Namespace missing
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("""
xsave(
  build({}, i),
  '{}_taz',
  namespace:taz)""".format(schema, url))

    # Public namespace
    scidb_con2.iquery("""
xsave(
  build({}, i),
  '{}_public')""".format(schema, url))
    scidb_con2.iquery("xinput('{}_public')".format(url))

    # w/o update permission
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery("""
xsave(
  build({}, i),
  '{}_foo',
  namespace:{})""".format(schema, url, test_ns))

    # w/ update permission
    scidb_con.set_role_permissions(test_user, "'namespace'", test_ns, 'u')
    scidb_con2 = scidbpy.connect(**con_args)
    scidb_con2.iquery("""
xsave(
  build({}, i),
  '{}_foo',
  namespace:{})""".format(schema, url, test_ns))

    # w/o read permission
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery("xinput('{}_foo')".format(url))

    # w/ read permission
    scidb_con.set_role_permissions(test_user, "'namespace'", test_ns, 'ru')
    scidb_con2 = scidbpy.connect(**con_args)
    scidb_con2.iquery("xinput('{}_foo')".format(url))

    # Cleanup
    scidb_con.drop_user(test_user)
    scidb_con.drop_namespace(test_ns)


def test_missing_index(scidb_con):
    url = '{}/xsave/missing_index'.format(test_url)
    schema = '<v:int64> [i=0:19:0:1]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    query = """
xsave(
  filter(
    build({}, i * 10),
    i < 10),
  '{}', update:true)""".format(schema, url)

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        {'i': range(20),
         'v': (float(i * 10) if i < 10 else float(i)
               for i in range(20))})
    pandas.testing.assert_frame_equal(array, array_gold)

    # Delete Index
    scidbbridge.driver.Driver.delete(url + '/index/0')

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)
    with pytest.raises(AssertionError):
        pandas.testing.assert_frame_equal(array, array_gold)
    pandas.testing.assert_frame_equal(
        array, pandas.DataFrame({'i': range(10),
                                 'v': (float(i * 10) for i in range(10))}))


def test_missing_index_big(scidb_con):
    url = '{}/xsave/missing_index_big'.format(test_url)
    schema = '<v:int64> [i=0:19:0:1; j=0:9:0:1]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i * j),
  '{}', index_split:100)""".format(schema, url))

    query = """
xsave(
  filter(
    build({}, i * 10),
    i < 10 and j < 5),
  '{}', update:true)""".format(schema, url)

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j, float(i * 10) if i < 10 and j < 5 else float(i * j))
              for i in range(20) for j in range(10)],
        columns=('i', 'j', 'v'))
    pandas.testing.assert_frame_equal(array, array_gold)

    # Delete From Index
    scidbbridge.driver.Driver.delete(url + '/index/0')

    # Update
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)


def test_missing_chunks(scidb_con):
    url = '{}/xsave/missing_chunks'.format(test_url)
    schema = '<v:int64> [i=0:19:0:1]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    query = """
xsave(
  filter(
    build({}, i * 10),
    i >= 10 and i < 15),
  '{}', update:true)""".format(schema, url)

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        {'i': range(20),
         'v': (float(i) if i < 10 or i >= 15 else float(i * 10)
               for i in range(20))})
    pandas.testing.assert_frame_equal(array, array_gold)

    scidbbridge.driver.Driver.delete(url + '/chunks/c_10')

    # Update
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)


def test_wrong_index(scidb_con):
    url = '{}/xsave/wrong_index'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    schema_build = schema.replace(', w:string', '')

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}')""".format(schema_build, url))

    query = """
xsave(
  filter(
    apply(
      build({}, i * 10),
      w, string(v)),
    i < 10 and j < 5),
  '{}', update:true)""".format(schema_build, url)

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j) + ((float(i * 10), str(i * 10)) if i < 10 and j < 5
                        else (float(i * j), str(i * j)))
              for i in range(20)
              for j in range(10)],
        columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    ar = scidbbridge.Array(url)
    index_gold = ar.read_index()
    index_url = '{}/index/0'.format(url)

    # ---
    # Add Column to Index
    index = index_gold.copy(True)
    index['k'] = index['i']
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Remove Column from Index
    index = index_gold.copy(True)
    index = index.drop(['j'], axis=1)
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Change Column to String in Index
    index = index_gold.copy(True)
    index['j'] = 'foo'
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Index Not Compressed
    index = index_gold.copy(True)
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Write Text File as Index
    parts = urllib.parse.urlparse(index_url)
    if parts.scheme == 's3':
        bucket = parts.netloc
        key = parts.path[1:]
        scidbbridge.driver.Driver.s3_client().put_object(
            Body="foo", Bucket=bucket, Key=key)
    elif parts.scheme == 'file':
        path = os.path.join(parts.netloc, parts.path)
        with open(path, 'w') as f:
            f.write("foo")
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


def test_wrong_chunk(scidb_con):
    url = '{}/xsave/wrong_chunk'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    schema_build = schema.replace(', w:string', '')

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}')""".format(schema_build, url))

    query = """
xsave(
  filter(
    apply(
      build({}, i * 10),
      w, string(v)),
    i < 10 and j < 5),
  '{}', update:true)""".format(schema_build, url)

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j) + ((float(i * 10), str(i * 10)) if i < 10 and j < 5
                        else (float(i * j), str(i * j)))
              for i in range(20)
              for j in range(10)],
        columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    ar = scidbbridge.Array(url)
    ch = ar.get_chunk(0, 0)
    chunk_gold = ch.to_pandas()
    chunk_url = '{}/chunks/c_0_0'.format(url)

    # ---
    # Add Column to Chunk
    chunk = chunk_gold.copy(True)
    chunk['x'] = chunk['v']
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Remove Column from Chunk
    chunk = chunk_gold.copy(True)
    chunk = chunk.drop(['w'], axis=1)
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Change Column to String in Chunk
    chunk = chunk_gold.copy(True)
    chunk['v'] = 'foo'
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Chunk Compressed
    chunk = chunk_gold.copy(True)
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Write Text File as Chunk
    parts = urllib.parse.urlparse(chunk_url)
    if parts.scheme == 's3':
        bucket = parts.netloc
        key = parts.path[1:]
        scidbbridge.driver.Driver.s3_client().put_object(
            Body="foo", Bucket=bucket, Key=key)
    elif parts.scheme == 'file':
        path = os.path.join(parts.netloc, parts.path)
        with open(path, 'w') as f:
            f.write("foo")
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


def test_wrong_chunk_compressed(scidb_con):
    url = '{}/xsave/wrong_chunk_compressed'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    schema_build = schema.replace(', w:string', '')

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}', compression:'gzip')""".format(schema_build, url))

    query = """
xsave(
  filter(
    apply(
      build({}, i * 10),
      w, string(v)),
    i < 10 and j < 5),
  '{}', update:true)""".format(schema_build, url)

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j) + ((float(i * 10), str(i * 10)) if i < 10 and j < 5
                        else (float(i * j), str(i * j)))
              for i in range(20)
              for j in range(10)],
        columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    ar = scidbbridge.Array(url)
    ch = ar.get_chunk(0, 0)
    chunk_gold = ch.to_pandas()
    chunk_url = '{}/chunks/c_0_0'.format(url)

    # ---
    # Add Column to Chunk
    chunk = chunk_gold.copy(True)
    chunk['x'] = chunk['v']
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Remove Column from Chunk
    chunk = chunk_gold.copy(True)
    chunk = chunk.drop(['w'], axis=1)
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Change Column to String in Chunk
    chunk = chunk_gold.copy(True)
    chunk['v'] = 'foo'
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Chunk Not Compressed
    chunk = chunk_gold.copy(True)
    chunk_table = pyarrow.Table.from_pandas(chunk)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Write Text File as Chunk
    parts = urllib.parse.urlparse(chunk_url)
    if parts.scheme == 's3':
        bucket = parts.netloc
        key = parts.path[1:]
        scidbbridge.driver.Driver.s3_client().put_object(
            Body="foo", Bucket=bucket, Key=key)
    elif parts.scheme == 'file':
        path = os.path.join(parts.netloc, parts.path)
        with open(path, 'w') as f:
            f.write("foo")
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


def test_wrong_metadata(scidb_con):
    url = '{}/xsave/wrong_metadata'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    schema_build = schema.replace(', w:string', '')

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}')""".format(schema_build, url))

    query = """
xsave(
  filter(
    apply(
      build({}, i * 10),
      w, string(v)),
    i < 10 and j < 5),
  '{}', update:true)""".format(schema_build, url)

    # Update
    scidb_con.iquery(query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j) + ((float(i * 10), str(i * 10)) if i < 10 and j < 5
                        else (float(i * j), str(i * j)))
              for i in range(20)
              for j in range(10)],
        columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    metadata_url = url + '/metadata'
    metadata_gold = scidbbridge.driver.Driver.read_metadata(url)
    metadata_gold_dict = scidbbridge.Array.metadata_from_string(metadata_gold)
    metadata_keys = ('attribute',
                     'compression',
                     'format',
                     'index_split',
                     'namespace',
                     'schema',
                     'version')

    # ---
    # Delete Metadata
    scidbbridge.driver.Driver.delete(metadata_url)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Delete Metadata Field
    for key in metadata_keys:
        metadata = metadata_gold_dict.copy()
        del metadata[key]
        save_metadata(metadata_url, metadata_dict2text(metadata))
        with pytest.raises(requests.exceptions.HTTPError):
            scidb_con.iquery(query)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Add Metadata Field
    metadata = metadata_gold_dict.copy()
    metadata['foo'] = 'bar'
    save_metadata(metadata_url, metadata_dict2text(metadata))
    scidb_con.iquery(query)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Wrong Metadata Field Value
    for key in metadata_keys:
        metadata = metadata_gold_dict.copy()
        metadata[key] = 'foo'
        save_metadata(metadata_url, metadata_dict2text(metadata))
        if key != 'namespace':
            with pytest.raises(requests.exceptions.HTTPError):
                scidb_con.iquery(query)
        else:
            scidb_con.iquery(query)
            array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
            array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
            pandas.testing.assert_frame_equal(array, array_gold)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Misstyped Type in Metadata Schema Value
    metadata = metadata_gold_dict.copy()
    metadata['schema'] = metadata['schema'].replace('int64', 'unt64')
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Wrong Type in Metadata Schema Value
    metadata = metadata_gold_dict.copy()
    metadata['schema'] = metadata['schema'].replace('string', 'int64')
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Wrong Syntax in Metadata Schema Value
    metadata = metadata_gold_dict.copy()
    metadata['schema'] = metadata['schema'].replace(']', '')
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Write Text File as Metadata
    save_metadata(metadata_url, "foo")
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


@pytest.mark.skipif(not security_enable,
                    reason="Requires SciDB Enterprise Edition")
def test_wrong_namespace(scidb_con):
    # Setup
    scidb_con.create_user(test_user, test_password_hash)

    con_args = {'scidb_url': scidb_url,
                'scidb_auth': (test_user, test_password),
                'verify': False}
    scidb_con2 = scidbpy.connect(**con_args)

    url = '{}/xsave/wrong_namespace'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    schema_build = schema.replace(', w:string', '')

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}')""".format(schema_build, url))

    query = """
xsave(
  filter(
    apply(
      build({}, i * 10),
      w, string(v)),
    i < 10 and j < 5),
  '{}', update:true)""".format(schema_build, url)

    # Update
    scidb_con2.iquery(query)

    # Input
    array = scidb_con2.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j) + ((float(i * 10), str(i * 10)) if i < 10 and j < 5
                        else (float(i * j), str(i * j)))
              for i in range(20)
              for j in range(10)],
        columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    metadata_url = url + '/metadata'
    metadata_gold = scidbbridge.driver.Driver.read_metadata(url)
    metadata_gold_dict = scidbbridge.Array.metadata_from_string(metadata_gold)
    metadata_keys = ('attribute',
                     'compression',
                     'format',
                     'index_split',
                     'namespace',
                     'schema',
                     'version')

    # ---
    # Wrong Namespace in Metadata
    metadata = metadata_gold_dict.copy()
    metadata['namespace'] = test_ns
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery(query)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con2.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # Cleanup
    scidb_con.drop_user(test_user)


@pytest.mark.skipif(not test_url.startswith('file://'),
                    reason="Requires file:// URL")
def test_wrong_url(scidb_con):
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    query = """
xsave(
  apply(
    build(<v:int64> [i=0:19:0:5; j=0:9:0:5], i * j),
    w, string(v)),
  '{}')"""

    # Invalid URL
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format(''))

    # Invalid URL
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file'))

    # Invalid URL
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file//'))

    # Path Empty
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file://'))

    # Path Relative
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file://foo'))


def test_s3_server_side_encryption(scidb_con):
    url = '{}/xsave/s3_sse'.format(test_url)
    schema = '<v:int64> [i=0:19:0:5]'

    if not url.startswith('s3://'):
        return

    # Wrong usage
    que = "xsave(build({}, i), '{}', {{}})".format(schema, url)
    for arg in ('s3_sse',
                's3_sse:0'
                's3_sse:true',
                "s3_sse:''",
                "s3_sse:'0'",
                "s3_sse:'true'",
                "s3_sse:'foo'"):
        with pytest.raises(requests.exceptions.HTTPError):
            scidb_con.iquery(que.format(arg))

    # Corect usage
    que = "xsave(build({}, i), '{}{{}}', {{}})".format(schema, url)
    i = 0
    for arg in ("s3_sse:'NOT_SET'",
                "s3_sse:'AES256'",
                "s3_sse:'aws:kms'"):
        scidb_con.iquery(que.format(i, arg))
        i += 1


def test_end_chunk(scidb_con):
    url = '{}/xsave/end_chunk'.format(test_url)
    schema = '<x:int64>[i=-3:17:0:5; j=7:23:0:4; k=-100:-77:0:7]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i * j * k),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j', 'k']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j, k, float(i * j * k))
              for i in range(-3, 17 + 1)
              for j in range(7, 23 + 1)
              for k in range(-100, -77 + 1)],
        columns=('i', 'j', 'k', 'x'))
    pandas.testing.assert_frame_equal(array, array_gold)


def test_chunk_overlap(scidb_con):
    url = '{}/xsave/chunk_overlap'.format(test_url)

    schema = '<x:int64>[i=0:9:4:5]'
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave(build({}, i), '{}')".format(schema, url))

    schema = '<x:int64>[i=0:9:0:5; j=-100:99:2:10]'
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave(build({}, i), '{}')".format(schema, url))

    schema = '<x:int64>[i=0:9:0:5; j=-100:99:1:10; k=0:9:0:5]'
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xsave(build({}, i), '{}')".format(schema, url))


def test_not_empty(scidb_con):
    url = '{}/xsave/not_empty'.format(test_url)
    schema = '<v:int64> [i=0:19:0:5; j=0:9:0:5]'

    # Store
    xsave_query = """
xsave(
  build({}, i * j),
  '{}')""".format(schema, url)
    scidb_con.iquery(xsave_query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j, float(i * j))
              for i in range(20)
              for j in range(10)],
        columns=('i', 'j', 'v'))
    pandas.testing.assert_frame_equal(array, array_gold)

    # Delete Metadata
    scidbbridge.driver.Driver.delete(url + '/metadata')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(xsave_query.replace('i * j', 'i + j'))


@pytest.mark.skipif(not security_enable,
                    reason="Requires SciDB Enterprise Edition")
def test_create_path(scidb_con):
    # Setup
    scidb_con.create_user(test_user, test_password_hash)

    con_args = {'scidb_url': scidb_url,
                'scidb_auth': (test_user, test_password),
                'verify': False}
    scidb_con2 = scidbpy.connect(**con_args)

    path = 'xsave/path'
    url = f'{test_url}/{path}/foo/bar'.format(test_url)
    schema = '<v:int64> [i=0:9:0:5; j=10:19:0:5]'

    # Queries
    xsave_query = """
xsave(
  build({}, i * j),
  '{}')""".format(schema, url)
    xremove_query = f"xremove('{url}')"

    # # As regular user
    scidb_con2 = scidbpy.connect(**con_args)
    scidb_con2.iquery(xsave_query)
    scidb_con2.iquery(xremove_query)

    # FS Cleanup
    parts = urllib.parse.urlparse(test_url)
    if parts.scheme == 'file':
        fs_path = os.path.join(parts.path, 'xsave/path')
        try:
            shutil.rmtree(fs_path)
        except PermissionError:
            pass
        except FileNotFoundError:
            pass

    # As root
    scidb_con.iquery(xsave_query)
    scidb_con.iquery(xremove_query)

    # Cleanup
    scidb_con.drop_user(test_user)


def test_list_operators(scidb_con):
    url = '{}/xsave/list_operators'.format(test_url)

    # Store
    xsave_query = """
xsave(
  list('operators'),
  '{}')""".format(url)
    scidb_con.iquery(xsave_query)

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)



# BEGIN_COPYRIGHT
#
# Copyright (C) 2020-2023 Paradigm4 Inc.
# All Rights Reserved.
#
# scidbbridge is a plugin for SciDB, an Open Source Array DBMS
# maintained by Paradigm4. See http://www.paradigm4.com/
#
# scidbbridge is free software: you can redistribute it and/or modify
# it under the terms of the AFFERO GNU General Public License as
# published by the Free Software Foundation.
#
# scidbbridge is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY
# KIND, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
# NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See the
# AFFERO GNU General Public License for the complete license terms.
#
# You should have received a copy of the AFFERO GNU General Public
# License along with scidbbridge. If not, see
# <http://www.gnu.org/licenses/agpl-3.0.html>
#
# END_COPYRIGHT

import itertools
import numpy
import pandas
import pyarrow
import pytest
import requests
import scidbbridge.coord
import urllib

from conftest import *


@pytest.mark.parametrize('chunk_size', (5, 10, 20))
def test_one_dim_one_attr(scidb_con, chunk_size):
    url = '{}/xinput/one_dim_one_attr_{}'.format(test_url, chunk_size)
    schema = '<v:int64> [i=0:19:0:{}]'.format(chunk_size)

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    pandas.testing.assert_frame_equal(
        array,
        pandas.DataFrame({'i': range(20), 'v': numpy.arange(0.0, 20.0)}))


@pytest.mark.parametrize('chunk_size', (5, 10, 20))
def test_multi_attr(scidb_con, chunk_size):
    url = '{}/xinput/multi_attr_{}'.format(test_url, chunk_size)
    schema = '<v:int64, w:int64> [i=0:19:0:{}]'.format(chunk_size)

    # Store
    scidb_con.iquery("""
xsave(
  redimension(
    apply(
      build({}, i),
      w, v * v),
    {}),
  '{}')""".format(schema.replace(', w:int64', ''),
                  schema,
                  url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    pandas.testing.assert_frame_equal(
        array,
        pandas.DataFrame({'i': range(20),
                          'v': numpy.arange(0.0, 20.0),
                          'w': (numpy.arange(0.0, 20.0) *
                                numpy.arange(0.0, 20.0))}))


@pytest.mark.parametrize('dim_start, dim_end, chunk_size',
                         ((s, e, c)
                          for s in (-21, -13, -10)
                          for e in (10, 16, 19)
                          for c in (5, 7, 10, 20)))
def test_multi_dim(scidb_con, dim_start, dim_end, chunk_size):
    url = '{}/xinput/multi_dim_{}_{}_{}'.format(
        test_url, dim_start, dim_end, chunk_size)
    schema = '<v:int64> [i={s}:{e}:0:{c}; j=-15:14:0:{c}]'.format(
        s=dim_start, e=dim_end - 1, c=chunk_size)

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    i_lst = []
    j_lst = []
    v_lst = []
    for i in range(dim_start, dim_end):
        for j in range(-15, 15):
            i_lst.append(i)
            j_lst.append(j)
            v_lst.append(float(i))

    pandas.testing.assert_frame_equal(array,
                                      pandas.DataFrame({'i': i_lst,
                                                        'j': j_lst,
                                                        'v': v_lst}))


@pytest.mark.parametrize(
    'type_name, is_null, type_numpy, chunk_size',
    ((t, n, p, c)
     for (t, n, p) in itertools.chain(
             ((t, n, object)
              for n in (True, False)
              for t in ('binary', 'string', 'char')),

             (('datetime', n, None)
              for n in (True, False)),

             (('bool', ) + p
              for p in ((True, object),
                        (False, bool))),

             ((t, n, tn)
              for (t, tn) in (('float', numpy.float32),
                              ('double', numpy.float64))
              for n in (True, False)),

             (('{}int{}'.format(g, s),
               n,
               numpy.dtype(
                   'float{}'.format(min(s * 2, 64)) if n
                   else '{}int{}'.format(g, s)).type)
              for g in ('', 'u')
              for s in (8, 16, 32, 64)
              for n in (True, False)))
     for c in (5, 10, 20)))
def test_type(scidb_con, type_name, is_null, type_numpy, chunk_size):
    max_val = 20
    url = '{}/xinput/type_{}_{}_{}'.format(
        test_url, type_name, is_null, chunk_size)
    schema = '<v:{} {}NULL> [i=0:{}:0:{}]'.format(
        type_name, '' if is_null else 'NOT ', max_val - 1, chunk_size)

    # Store
    if type_name == 'binary':
        que = scidb_con.input(
            schema.replace('binary NULL', 'binary NOT NULL'),
            upload_data=numpy.array([bytes([i]) for i in range(max_val)],
                                    dtype='object'))
        if is_null:
            que = que.redimension(schema)
    else:
        que = scidb_con.build(schema, '{}(i)'.format(type_name))
    que.xsave("'{}'".format(url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    if type_name in ('binary', 'char'):
        v = pandas.Series((bytes([i]) if i != 0 or type_name == 'binary'
                           else bytes()
                           for i in range(max_val)),
                          dtype=object)
    elif type_name == 'string':
        v = pandas.Series((str(i) for i in range(max_val)),
                          dtype=object)
    elif type_name.startswith('datetime'):
        if is_null:
            dtype = 'datetime64[ns]'
        else:
            dtype = 'datetime64[s]'
        v = pandas.Series((pandas.Timestamp(i * 10 ** 9)
                           for i in range(max_val)),
                          dtype=dtype)
    elif type_name == 'bool':
        if is_null:
            v = pandas.Series((bool(i) for i in range(max_val)),
                              dtype=object)
        else:
            v = pandas.Series((bool(i) for i in range(max_val)))
    else:
        v = type_numpy(range(max_val))

    pandas.testing.assert_frame_equal(
        array,
        pandas.DataFrame({'i': range(max_val), 'v': v}))


# Test for Empty Cells
@pytest.mark.parametrize('dim_start, dim_end, chunk_size',
                         ((s, e, c)
                          for s in (-13, -11, -7)
                          for e in (11, 13, 17)
                          for c in (3, 7, 11)))
def test_filter_before(scidb_con, dim_start, dim_end, chunk_size):
    url = '{}/xinput/filter_before_{}_{}_{}'.format(
        test_url, dim_start, dim_end, chunk_size)
    schema = '<v:int64> [i={s}:{e}:0:{c}; j=-11:13:0:{c}]'.format(
        s=dim_start, e=dim_end - 1, c=chunk_size)

    # Store
    scidb_con.iquery("""
xsave(
  filter(build({}, i), i % 3 = 0 and i > 7),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    i_lst = []
    j_lst = []
    v_lst = []
    for i in range(dim_start, dim_end):
        for j in range(-11, 14):
            if i % 3 == 0 and i > 7:
                i_lst.append(i)
                j_lst.append(j)
                v_lst.append(float(i))

    pandas.testing.assert_frame_equal(array,
                                      pandas.DataFrame({'i': i_lst,
                                                        'j': j_lst,
                                                        'v': v_lst}))


@pytest.mark.parametrize('dim_start, dim_end, chunk_size',
                         ((s, e, c)
                          for s in (-13, -11, -7)
                          for e in (11, 13, 17)
                          for c in (3, 7, 11)))
def test_filter_after(scidb_con, dim_start, dim_end, chunk_size):
    url = '{}/xinput/filter_after_{}_{}_{}'.format(
        test_url, dim_start, dim_end, chunk_size)
    schema = '<v:int64> [i={s}:{e}:0:{c}; j=-11:13:0:{c}]'.format(
        s=dim_start, e=dim_end - 1, c=chunk_size)

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("""
filter(
  xinput(
    '{}'),
  i % 3 = 0 and i > 7)""".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    i_lst = []
    j_lst = []
    v_lst = []
    for i in range(dim_start, dim_end):
        for j in range(-11, 14):
            if i % 3 == 0 and i > 7:
                i_lst.append(i)
                j_lst.append(j)
                v_lst.append(float(i))

    pandas.testing.assert_frame_equal(array,
                                      pandas.DataFrame({'i': i_lst,
                                                        'j': j_lst,
                                                        'v': v_lst}))


def test_nulls(scidb_con):
    url = '{}/xinput/nulls'.format(test_url)
    schema = '<v:int64> [i=0:99:0:5]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, iif(i % 2 = 0, i, missing(i))),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    i_lst = []
    v_lst = []
    for i in range(100):
        i_lst.append(i)
        if i % 2 == 0:
            v_lst.append(float(i))
        else:
            v_lst.append(numpy.nan)

    pandas.testing.assert_frame_equal(array,
                                      pandas.DataFrame({'i': i_lst,
                                                        'v': v_lst}))


def test_chunk_index(scidb_con):
    size = 300
    url = '{}/xinput/chunk_index'.format(test_url)
    schema = '<v:int64> [i=0:{}:0:5]'.format(size - 1)

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    pandas.testing.assert_frame_equal(
        array,
        pandas.DataFrame({'i': range(size),
                          'v': numpy.arange(0.0, float(size))}))


# Test with Different Cache Sizes
@pytest.mark.parametrize('cache_size', (None, 4000, 3000, 0))
def test_cache(scidb_con, cache_size):
    url = '{}/xinput/cache-{}'.format(test_url, cache_size)
    schema = '<v:int64 not null, w:int64 not null> [i=0:999:0:100]'

    # Store
    scidb_con.iquery("""
xsave(
  redimension(
    filter(
      apply(
        build({}, i),
        w, i * i),
      i % 100 < 80 or i >= 800),
    {}),
  '{}')""".format(schema.replace(', w:int64 not null', ''),
                  schema,
                  url))

    # Input
    que = "xinput('{}'{})".format(
        url,
        '' if cache_size is None else ', cache_size:{}'.format(cache_size))

    if cache_size == 3000:
        with pytest.raises(requests.exceptions.HTTPError):
            array = scidb_con.iquery(que, fetch=True)
    else:
        array = scidb_con.iquery(que, fetch=True)
        array = array.sort_values(by=['i']).reset_index(drop=True)

        pandas.testing.assert_frame_equal(
            array,
            pandas.DataFrame(data=((i, i, i * i)
                                   for i in range(1000)
                                   if (i % 100 < 80 or i >= 800)),
                             columns=('i', 'v', 'w')))


# Test with Multiple Arrow Chunks per File
def test_arrow_chunk(scidb_con):
    prefix = 'xinput/arrow_chunk'
    url = '{}/{}'.format(test_url, prefix)
    schema = '<v:int64> [i=0:999:0:1000]'

    # Store
    # if url.startswith('s3://'):
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # Re-write one SciDB Chunk file to use multiple Arrow Chunks
    parts = urllib.parse.urlparse(test_url)
    if parts.scheme == 's3':
        s3_key = '{}/{}/chunks/c_0'.format(parts.path[1:], prefix)
        obj = s3_con.get_object(Bucket=parts.netloc, Key=s3_key)
        reader = pyarrow.RecordBatchStreamReader(
            pyarrow.input_stream(pyarrow.py_buffer(obj['Body'].read()),
                                 compression='lz4'))
    elif parts.scheme == 'file':
        fn = '{}/{}/chunks/c_0'.format(parts.path, prefix)
        reader = pyarrow.RecordBatchStreamReader(
            pyarrow.input_stream(fn, compression='lz4'))

    tbl = reader.read_all()

    if url.startswith('s3://'):
        sink = pyarrow.BufferOutputStream()
        writer = pyarrow.ipc.RecordBatchStreamWriter(sink, tbl.schema)
    elif url.startswith('file://'):
        writer = pyarrow.ipc.RecordBatchStreamWriter(fn, tbl.schema)

    batches = tbl.to_batches(max_chunksize=200)  # 1000 / 200 = 5 chunks
    writer.write_table(pyarrow.Table.from_batches(batches))
    writer.close()

    if parts.scheme == 's3':
        s3_con.put_object(Body=sink.getvalue().to_pybytes(),
                          Bucket=parts.netloc,
                          Key=s3_key)

    # Input
    que = "xinput('{}')".format(url)

    with pytest.raises(requests.exceptions.HTTPError):
        array = scidb_con.iquery(que, fetch=True)


def test_missing_index(scidb_con):
    url = '{}/xinput/missing_index'.format(test_url)
    schema = '<v:int64> [i=0:19:0:1]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    array_gold = pandas.DataFrame({'i': range(20), 'v': map(float, range(20))})
    pandas.testing.assert_frame_equal(array, array_gold)

    scidbbridge.driver.Driver.delete(url + '/index/0')

    # Empty Array
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    with pytest.raises(AssertionError):
        pandas.testing.assert_frame_equal(array, array_gold)
    assert len(array) == 0


def test_missing_index_big(scidb_con):
    url = '{}/xinput/missing_index_big'.format(test_url)
    schema = '<v:int64> [i=0:19:0:1; j=0:9:0:1]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i * j),
  '{}', index_split:100)""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        data=[(i, j, float(i * j)) for i in range(20) for j in range(10)],
        columns=('i', 'j', 'v'))
    pandas.testing.assert_frame_equal(array, array_gold)

    scidbbridge.driver.Driver.delete(url + '/index/0')

    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)


def test_missing_chunks(scidb_con):
    url = '{}/xinput/missing_chunks'.format(test_url)
    schema = '<v:int64> [i=0:19:0:1]'

    # Store
    scidb_con.iquery("""
xsave(
  build({}, i),
  '{}')""".format(schema, url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    array_gold = pandas.DataFrame({'i': range(20), 'v': map(float, range(20))})
    pandas.testing.assert_frame_equal(array, array_gold)

    scidbbridge.driver.Driver.delete(url + '/chunks/c_0')

    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)


def test_wrong_index(scidb_con):
    url = '{}/xinput/wrong_index'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}')""".format(schema.replace(', w:string', ''),
                  url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(data=[(i, j, float(i * j), str(i * j))
                                        for i in range(20)
                                        for j in range(10)],
                                  columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    ar = scidbbridge.Array(url)
    index_gold = ar.read_index()
    index_url = '{}/index/0'.format(url)

    # ---
    # Add Column to Index
    index = index_gold.copy(True)
    index['k'] = index['i']
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Remove Column from Index
    index = index_gold.copy(True)
    index = index.drop(['j'], axis=1)
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Change Column to String in Index
    index = index_gold.copy(True)
    index['j'] = 'foo'
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema, compression='gzip')
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Index Not Compressed
    index = index_gold.copy(True)
    index_table = pyarrow.Table.from_pandas(index)
    scidbbridge.driver.Driver.write_table(
        index_table, index_url, index_table.schema)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Write Text File as Index
    parts = urllib.parse.urlparse(index_url)
    if parts.scheme == 's3':
        bucket = parts.netloc
        key = parts.path[1:]
        scidbbridge.driver.Driver.s3_client().put_object(
            Body="foo", Bucket=bucket, Key=key)
    elif parts.scheme == 'file':
        path = os.path.join(parts.netloc, parts.path)
        with open(path, 'w') as f:
            f.write("foo")
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ar.write_index(index_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


@pytest.mark.parametrize('compression', (None, 'gzip', 'lz4'))
def test_wrong_chunk(scidb_con, compression):
    compression_human = compression if compression is not None else 'none'
    url = '{}/xinput/wrong_chunk/{}'.format(test_url, compression_human)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    schema_raw = [('v', pyarrow.int64(), True),
                  ('w', pyarrow.string(), True),
                  ('@delta', pyarrow.int64(), False)]

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}', compression:'{}')""".format(schema.replace(', w:string', ''),
                                    url,
                                    compression_human))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(data=[(i, j, float(i * j), str(i * j))
                                        for i in range(20)
                                        for j in range(10)],
                                  columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    ar = scidbbridge.Array(url)
    ch = ar.get_chunk(0, 0)
    chunk_gold = ch.to_pandas()
    chunk_url = '{}/chunks/c_0_0'.format(url)

    # ---
    # Add Column to Chunk
    chunk = chunk_gold.copy(True)
    chunk['x'] = chunk['v']
    chunk['@delta'] = scidbbridge.coord.coord2delta(
        chunk, ch.array.schema.dims, ch.coords)
    schema_arrow = pyarrow.schema(
        schema_raw[:2] +
        [('x', pyarrow.int64(), True)] +
        schema_raw[2:])
    chunk_table = pyarrow.Table.from_pandas(chunk, schema_arrow)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema, compression=compression)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Remove Column from Chunk
    chunk = chunk_gold.copy(True)
    chunk = chunk.drop(['w'], axis=1)
    chunk['@delta'] = scidbbridge.coord.coord2delta(
        chunk, ch.array.schema.dims, ch.coords)
    schema_arrow = pyarrow.schema(schema_raw[:1] + schema_raw[2:])
    chunk_table = pyarrow.Table.from_pandas(chunk, schema_arrow)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema, compression=compression)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Change Column to String in Chunk
    chunk = chunk_gold.copy(True)
    chunk['v'] = 'foo'
    chunk['@delta'] = scidbbridge.coord.coord2delta(
        chunk, ch.array.schema.dims, ch.coords)
    schema_arrow = pyarrow.schema(
        [('v', pyarrow.string(), True)] +
        schema_raw[1:])
    chunk_table = pyarrow.Table.from_pandas(chunk, schema_arrow)
    scidbbridge.driver.Driver.write_table(
        chunk_table, chunk_url, chunk_table.schema, compression=compression)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Chunk w/ Different Compression
    chunk = chunk_gold.copy(True)
    chunk['@delta'] = scidbbridge.coord.coord2delta(
        chunk, ch.array.schema.dims, ch.coords)
    schema_arrow = pyarrow.schema(schema_raw)
    chunk_table = pyarrow.Table.from_pandas(chunk, schema_arrow)
    not_compression = None if compression else 'gzip'
    scidbbridge.driver.Driver.write_table(chunk_table,
                                          chunk_url,
                                          chunk_table.schema,
                                          compression=not_compression)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Write Text File as Chunk
    parts = urllib.parse.urlparse(chunk_url)
    if parts.scheme == 's3':
        bucket = parts.netloc
        key = parts.path[1:]
        scidbbridge.driver.Driver.s3_client().put_object(
            Body="foo", Bucket=bucket, Key=key)
    elif parts.scheme == 'file':
        path = os.path.join(parts.netloc, parts.path)
        with open(path, 'w') as f:
            f.write("foo")
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    ch.from_pandas(chunk_gold)
    ch.save()
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


def test_wrong_metadata(scidb_con):
    url = '{}/xinput/wrong_metadata'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}')""".format(schema.replace(', w:string', ''),
                  url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(data=[(i, j, float(i * j), str(i * j))
                                        for i in range(20)
                                        for j in range(10)],
                                  columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    metadata_url = url + '/metadata'
    metadata_gold = scidbbridge.driver.Driver.read_metadata(url)
    metadata_gold_dict = scidbbridge.Array.metadata_from_string(metadata_gold)
    metadata_keys = ('attribute',
                     'compression',
                     'format',
                     'index_split',
                     'namespace',
                     'schema',
                     'version')

    # ---
    # Delete Metadata
    scidbbridge.driver.Driver.delete(metadata_url)
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Delete Metadata Field
    for key in metadata_keys:
        metadata = metadata_gold_dict.copy()
        del metadata[key]
        save_metadata(metadata_url, metadata_dict2text(metadata))
        with pytest.raises(requests.exceptions.HTTPError):
            scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Add Metadata Field
    metadata = metadata_gold_dict.copy()
    metadata['foo'] = 'bar'
    save_metadata(metadata_url, metadata_dict2text(metadata))
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Wrong Metadata Field Value
    for key in metadata_keys:
        metadata = metadata_gold_dict.copy()
        metadata[key] = 'foo'
        save_metadata(metadata_url, metadata_dict2text(metadata))
        if key != 'namespace':
            with pytest.raises(requests.exceptions.HTTPError):
                scidb_con.iquery("xinput('{}')".format(url), fetch=True)
        else:
            array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
            array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
            pandas.testing.assert_frame_equal(array, array_gold)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Misstyped Type in Metadata Schema Value
    metadata = metadata_gold_dict.copy()
    metadata['schema'] = metadata['schema'].replace('int64', 'unt64')
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Wrong Type in Metadata Schema Value
    metadata = metadata_gold_dict.copy()
    metadata['schema'] = metadata['schema'].replace('string', 'int64')
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Wrong Syntax in Metadata Schema Value
    metadata = metadata_gold_dict.copy()
    metadata['schema'] = metadata['schema'].replace(']', '')
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Write Text File as Metadata
    save_metadata(metadata_url, "foo")
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


@pytest.mark.skipif(not security_enable,
                    reason="Requires SciDB Enterprise Edition")
def test_wrong_namespace(scidb_con):
    # Setup
    scidb_con.create_user(test_user, test_password_hash)

    con_args = {'scidb_url': scidb_url,
                'scidb_auth': (test_user, test_password),
                'verify': False}
    scidb_con2 = scidbpy.connect(**con_args)

    url = '{}/xinput/wrong_namespace'.format(test_url)
    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'

    # FS Init
    parts = urllib.parse.urlparse(test_url)
    if parts.scheme == 'file':
        fs_path = os.path.join(parts.path, 'xinput')
        if not os.path.exists(fs_path):
            os.makedirs(fs_path)

    # Store
    scidb_con2.iquery("""
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{}')""".format(schema.replace(', w:string', ''),
                  url))

    # Input
    array = scidb_con2.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(data=[(i, j, float(i * j), str(i * j))
                                        for i in range(20)
                                        for j in range(10)],
                                  columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    metadata_url = url + '/metadata'
    metadata_gold = scidbbridge.driver.Driver.read_metadata(url)
    metadata_gold_dict = scidbbridge.Array.metadata_from_string(metadata_gold)
    metadata_keys = ('attribute',
                     'compression',
                     'format',
                     'index_split',
                     'namespace',
                     'schema',
                     'version')

    # ---
    # Wrong Namespace in Metadata
    metadata = metadata_gold_dict.copy()
    metadata['namespace'] = 'foo'
    save_metadata(metadata_url, metadata_dict2text(metadata))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery("xinput('{}')".format(url), fetch=True)

    # Restore
    save_metadata(metadata_url, metadata_gold)
    array = scidb_con2.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)

    # ---
    # Cleanup
    scidb_con.drop_user(test_user)


@pytest.mark.skipif(not test_url.startswith('file://'),
                    reason="Requires file:// URL")
def test_wrong_url(scidb_con):
    query = "xinput('{}')"

    # Invalid URL
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format(''))

    # Invalid URL
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file'))

    # Invalid URL
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file//'))

    # Path Empty
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file://'))

    # Path Relative
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con.iquery(query.format('file://foo'))


@pytest.mark.skipif(not security_enable
                    or not test_url.startswith('file://'),
                    reason="Requires SciDB Enterprise Edition and file:// URL")
def test_io_paths_list(scidb_con):
    # Setup
    scidb_con.create_user(test_user, test_password_hash)

    con_args = {'scidb_url': scidb_url,
                'scidb_auth': (test_user, test_password),
                'verify': False}
    scidb_con2 = scidbpy.connect(**con_args)

    parts = urllib.parse.urlparse(test_url)
    url_prefix = 'file://'
    url_postfix = '/io_paths_list'
    url1 = '{}{}{}'.format(url_prefix, parts.path, url_postfix)
    url2 = '{}/var/lib'.format(url_prefix)
    url3 = '{}/non/existing'.format(url_prefix)

    schema = '<v:int64, w:string> [i=0:19:0:5; j=0:9:0:5]'
    query = """
xsave(
  apply(
    build({}, i * j),
    w, string(v)),
  '{{}}')""".format(schema.replace(', w:string', ''))

    # Store
    scidb_con2.iquery(query.format(url1))

    # Input
    array = scidb_con2.iquery("xinput('{}')".format(url1), fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    array_gold = pandas.DataFrame(data=[(i, j, float(i * j), str(i * j))
                                        for i in range(20)
                                        for j in range(10)],
                                  columns=('i', 'j', 'v', 'w'))
    pandas.testing.assert_frame_equal(array, array_gold)

    # Not Allowed
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery(query.format(url2))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery("xinput('{}')".format(url2), fetch=True)

    # Does Not Exist
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery(query.format(url3))
    with pytest.raises(requests.exceptions.HTTPError):
        scidb_con2.iquery("xinput('{}')".format(url3), fetch=True)

    # Cleanup
    scidb_con.drop_user(test_user)


def test_from_pandas(scidb_con):
    url = '{}/xinput/from_pandas'.format(test_url)
    schema = '<v:int64, w:int64> [i=0:19:0:10]'

    # Store
    scidb_con.iquery("""
xsave(
  apply(
    build({}, -i),
    w, -v * 10),
  '{}')""".format(schema.replace(', w:int64', ''), url))

    # Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)

    array_gold = pandas.DataFrame(
        {'i': range(20),
         'v': map(lambda i: -float(i), range(20)),
         'w': map(lambda i: float(i * 10), range(20))})
    pandas.testing.assert_frame_equal(array, array_gold)

    # Missing Columns
    chunk = scidbbridge.Array(url).get_chunk(10)
    with pytest.raises(Exception) as ex:
        chunk.from_pandas(
            pandas.DataFrame(((i, i * 10) for i in range(10, 20)),
                             columns=list('iw')))
    assert ('do not match array attributes plus dimensions count'
            in str(ex.value))

    # Wrong Columns
    chunk = scidbbridge.Array(url).get_chunk(10)
    with pytest.raises(Exception) as ex:
        chunk.from_pandas(
            pandas.DataFrame(((i, i * 10, -i) for i in range(10, 20)),
                             columns=list('iwx')))
    assert 'does not match array schema' in str(ex.value)

    # Swapped Columns
    chunk = scidbbridge.Array(url).get_chunk(10)
    chunk.from_pandas(
        pandas.DataFrame(((i, i * 10, -i) for i in range(10, 20)),
                         columns=list('iwv')))
    chunk.save()

    # Check Input
    array = scidb_con.iquery("xinput('{}')".format(url), fetch=True)
    array = array.sort_values(by=['i']).reset_index(drop=True)
    pandas.testing.assert_frame_equal(array, array_gold)


def test_save(scidb_con):
    url = '{}/xinput/save'.format(test_url)
    schema = '<v:int64> [i=0:19:0:5]'

    # Store
    scidb_con.iquery("xsave(build({}, i), '{}')".format(schema, url))

    # Save
    scidb_con.iquery("save(xinput('{}'), '/dev/null')".format(url))


def test_filter_read(scidb_con):
    url = '{}/xinput/filter_sparse'.format(test_url)
    schema = '<v:int64> [i=1:10:0:10; j=1:10:0:10]'

    # Store
    scidb_con.iquery("""
xsave(
  filter(build({}, i * j), v % 3 = 0),
  '{}')""".format(schema, url))

    # Input & Filter
    array = scidb_con.iquery("filter(xinput('{}'), i % 2 = 0)".format(url),
                             fetch=True)
    array = array.sort_values(by=['i', 'j']).reset_index(drop=True)

    i_lst = []
    j_lst = []
    v_lst = []
    for i in range(1, 11):
        for j in range(1, 11):
            if i * j % 3 == 0 and i % 2 == 0:
                i_lst.append(i)
                j_lst.append(j)
                v_lst.append(float(i * j))

    pandas.testing.assert_frame_equal(array,
                                      pandas.DataFrame({'i': i_lst,
                                                        'j': j_lst,
                                                        'v': v_lst}))



# BEGIN_COPYRIGHT
#
# Copyright (C) 2020-2023 Paradigm4 Inc.
# All Rights Reserved.
#
# scidbbridge is a plugin for SciDB, an Open Source Array DBMS
# maintained by Paradigm4. See http://www.paradigm4.com/
#
# scidbbridge is free software: you can redistribute it and/or modify
# it under the terms of the AFFERO GNU General Public License as
# published by the Free Software Foundation.
#
# scidbbridge is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY
# KIND, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
# NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See the
# AFFERO GNU General Public License for the complete license terms.
#
# You should have received a copy of the AFFERO GNU General Public
# License along with scidbbridge. If not, see
# <http://www.gnu.org/licenses/agpl-3.0.html>
#
# END_COPYRIGHT

import boto3
import os
import pyarrow
import pyarrow.parquet
import urllib.parse
import shutil


class Driver:
    default_format = 'arrow'
    default_compression = 'lz4'

    index_format = 'arrow'
    index_compression = 'lz4'

    _s3_client = None
    _s3_resource = None

    @staticmethod
    def s3_client():
        if Driver._s3_client is None:
            Driver._s3_client = boto3.client('s3')
        return Driver._s3_client

    @staticmethod
    def s3_resource():
        if Driver._s3_resource is None:
            Driver._s3_resource = boto3.resource('s3')
        return Driver._s3_resource

    @staticmethod
    def list(url):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            bucket = parts.netloc
            key = parts.path[1:] + '/'
            pages = Driver.s3_client().get_paginator(
                'list_objects_v2').paginate(Bucket=bucket, Prefix=key)
            for page in pages:
                if 'Contents' in page.keys():
                    for obj in page['Contents']:
                        yield 's3://{}/{}'.format(bucket, obj['Key'])

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path)
            for fn in os.listdir(path):
                if os.path.isfile(os.path.join(path, fn)):
                    yield 'file://' + os.path.join(path, fn)

        else:
            raise Exception('URL {} not supported'.format(url))

    @staticmethod
    def init_array(url):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            pass

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path)
            os.makedirs(path, exist_ok=True)
            os.mkdir(os.path.join(path, 'index'))
            os.mkdir(os.path.join(path, 'chunks'))

        else:
            raise Exception('URL {} not supported'.format(url))

    @staticmethod
    def read_metadata(url):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            bucket = parts.netloc
            key = parts.path[1:] + '/metadata'
            obj = Driver.s3_client().get_object(Bucket=bucket, Key=key)
            return obj['Body'].read().decode('utf-8')

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path, 'metadata')
            return open(path).read()

        else:
            raise Exception('URL {} not supported'.format(url))

    @staticmethod
    def write_metadata(url, metadata):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            bucket = parts.netloc
            key = parts.path[1:] + '/metadata'
            Driver.s3_client().put_object(Body=metadata,
                                          Bucket=bucket,
                                          Key=key)

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path, 'metadata')
            with open(path, 'w') as f:
                f.write(metadata)

        else:
            raise Exception('URL {} not supported'.format(url))

    @staticmethod
    def read_table(url,
                   format=default_format,
                   compression=default_compression):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            bucket = parts.netloc
            key = parts.path[1:]
            obj = Driver.s3_client().get_object(Bucket=bucket, Key=key)
            buf = pyarrow.py_buffer(obj['Body'].read())
            if format == 'arrow':
                strm = pyarrow.input_stream(buf,
                                            compression=compression)
                return pyarrow.RecordBatchStreamReader(strm).read_all()
            elif format == 'parquet':
                return pyarrow.parquet.read_table(buf)
            else:
                raise Exception('Format {} not supported'.format(format))

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path)
            if format == 'arrow':
                strm = pyarrow.input_stream(path, compression=compression)
                return pyarrow.RecordBatchStreamReader(strm).read_all()
            if format == 'parquet':
                return pyarrow.parquet.read_table(path)
            else:
                raise Exception('Format {} not supported'.format(format))

        else:
            raise Exception('URL {} not supported'.format(url))

    @staticmethod
    def write_table(table,
                    url,
                    schema,
                    format=default_format,
                    compression=default_compression):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            bucket = parts.netloc
            key = parts.path[1:]
            buf = pyarrow.BufferOutputStream()
            if format == 'arrow':
                stream = pyarrow.output_stream(buf, compression=compression)
                writer = pyarrow.RecordBatchStreamWriter(stream, schema)
                writer.write_table(table)
                writer.close()
                stream.close()
            elif format == 'parquet':
                pyarrow.parquet.write_table(
                    table, buf, compression=compression)
            else:
                raise Exception('Format {} not supported'.format(format))
            Driver.s3_client().put_object(Body=buf.getvalue().to_pybytes(),
                                          Bucket=bucket,
                                          Key=key)

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path)
            if format == 'arrow':
                stream = pyarrow.output_stream(path, compression=compression)
                writer = pyarrow.RecordBatchStreamWriter(stream, schema)
                writer.write_table(table)
                writer.close()
                stream.close()
            elif format == 'parquet':
                pyarrow.parquet.write_table(
                    table, path, compression=compression)
            else:
                raise Exception('Format {} not supported'.format(format))

        else:
            raise Exception('URL {} not supported'.format(url))

    @staticmethod
    def delete_all(url):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            bucket = parts.netloc
            key = parts.path[1:]
            Driver.s3_resource().Bucket(
                bucket).objects.filter(Prefix=key).delete()

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path)
            for fn in os.listdir(path):
                tfn = os.path.join(path, fn)
                if os.path.isdir(tfn):
                    shutil.rmtree(tfn)
                else:
                    os.unlink(tfn)

        else:
            raise Exception('URL {} not supported'.format(url))

    @staticmethod
    def delete(url):
        parts = urllib.parse.urlparse(url)

        # S3
        if parts.scheme == 's3':
            bucket = parts.netloc
            key = parts.path[1:]
            Driver.s3_client().delete_object(Bucket=bucket, Key=key)

        # File System
        elif parts.scheme == 'file':
            path = os.path.join(parts.netloc, parts.path)
            os.unlink(path)

        else:
            raise Exception('URL {} not supported'.format(url))



#!/bin/env python3

# BEGIN_COPYRIGHT
#
# Copyright (C) 2020-2023 Paradigm4 Inc.
# All Rights Reserved.
#
# scidbbridge is a plugin for SciDB, an Open Source Array DBMS
# maintained by Paradigm4. See http://www.paradigm4.com/
#
# scidbbridge is free software: you can redistribute it and/or modify
# it under the terms of the AFFERO GNU General Public License as
# published by the Free Software Foundation.
#
# scidbbridge is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY
# KIND, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
# NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See the
# AFFERO GNU General Public License for the complete license terms.
#
# You should have received a copy of the AFFERO GNU General Public
# License along with scidbbridge. If not, see
# <http://www.gnu.org/licenses/agpl-3.0.html>
#
# END_COPYRIGHT

import os
import pyarrow
import sys

from scidbbridge import Array, type_map_pyarrow
from scidbbridge.coord import coord2delta
from scidbbridge.driver import Driver


def wrong_arg():
    print("""Upgrade an existing Bridge Array from v1 to v3.

Usage:

{} URL""".format(os.path.basename(__file__)))
    sys.exit(2)


if len(sys.argv) != 2:
    wrong_arg()

url = sys.argv[1]

for index_url in Driver.list('{}/index'.format(url)):
    print('Fixing', index_url)

    # Read index with GZIP compression
    reader = Driver.create_reader(index_url, 'gzip')
    table = reader.read_all()

    # Fix nullness in Arrow schema
    schema = pyarrow.schema([(name, pyarrow.int64(), False)
                             for name in table.schema.names])
    table = table.cast(schema)

    # Re-write index with LZ4 compression
    sink = Driver.create_writer(index_url, table.schema, 'lz4')
    writer = next(sink)
    writer.write_table(table)
    sink.close()

array = Array(url)
old_compression = array.metadata['compression']

# Change None compression to LZ4
if old_compression is None:
    array.metadata['compression'] = 'lz4'
    Driver.write_metadata(url,
                          Array.metadata_to_string(array.metadata))

# Fixed schema
schema = pyarrow.schema(
    [(a.name, type_map_pyarrow[a.type_name], not a.not_null)
     for a in array.schema.atts] +
    [('@delta', pyarrow.int64(), False)])

for (_, pos) in array.read_index().iterrows():
    chunk = array.get_chunk(*pos.tolist())
    print('Fixing', chunk.url)

    # Read chunk
    reader = Driver.create_reader(chunk.url, old_compression)
    data = reader.read_all().to_pandas()

    # Convert coordinates to delta
    data['@delta'] = coord2delta(data, array.schema.dims, chunk.coords)

    # Prepare chunk data
    table = pyarrow.Table.from_pandas(data, schema)

    # Re-write index with LZ4 compression
    sink = Driver.create_writer(chunk.url,
                                table.schema,
                                array.metadata['compression'])
    writer = next(sink)
    writer.write_table(table)
    sink.close()



import itertools
import numpy
import pandas
import pyarrow
import sys

# import timeit

# from coord import *


# orig: chunk origin coordinates

# def coord2pos(coord, orig, chunk_size):
#     n_dims = len(coord)

#     if n_dims == 1:
#         pos = coord[0] - orig[0]
#     elif n_dims == 2:
#         pos = ((coord[0] - orig[0]) * chunk_size[1] +
#                (coord[1] - orig[1]))
#     else:
#         pos = 0
#         for i in range(n_dims):
#             pos *= chunk_size[i]
#             pos += coord[i] - orig[i]

#     return pos


# def coord2pos_all(data, dim_names, orig_names, chunk_size):
#     n_dims = len(dim_names)

#     if n_dims == 1:
#         res = data[dim_names[0]] - data[orig_names[0]]
#     elif n_dims == 2:
#         res = ((data[dim_names[0]] - data[orig_names[0]]) * chunk_size[1] +
#                data[dim_names[1]] - data[orig_names[1]])
#     else:
#         res = pandas.Series(numpy.zeros(len(data)), dtype=numpy.int64)
#         for i in range(n_dims):
#             res *= chunk_size[i]
#             res += data[dim_names[i]] - data[orig_names[i]]

#     return res


# def pos2coord(pos, orig, chunk_size):
#     coord = []
#     n_dims = len(orig)

#     if n_dims == 1:
#         coord.append(orig[0] + pos)                     # coord[0]
#     elif n_dims == 2:
#         coord.insert(0, orig[1] + pos % chunk_size[1])  # coord[1]
#         pos //= chunk_size[1]
#         coord.insert(0, orig[0] + pos)                  # coord[0]
#     else:
#         for i in range(n_dims - 1, -1, -1):
#             coord.insert(0, orig[i] + pos % chunk_size[i])
#             pos //= chunk_size[i]

#     return coord


# def pos2coord_all(data, pos_name, orig_names, chunk_size, res_dim_names):
#     n_dims = len(orig_names)
#     pos = data[pos_name]

#     if n_dims == 1:
#         res = (data[orig_names[0]] + pos, )
#     elif n_dims == 2:
#         coord1 = data[orig_names[1]] + data[pos_name] % chunk_size[1]
#         pos = pos // chunk_size[1]
#         coord0 = data[orig_names[0]] + pos
#         res = (coord0, coord1)
#     else:
#         res = []
#         for i in range(n_dims - 1, -1, -1):
#             res.insert(0, data[orig_names[i]] + pos % chunk_size[i])
#             pos = pos // chunk_size[i]

#     return pandas.DataFrame(dict(zip(res_dim_names, res)))


# def prev_delta2coord(prev_coord, delta, chunk_sizes):
#     """
#     >>> prev_delta2coord((1, 1, 0), 4, (4, 3, 2))
#     [2, 0, 0]
#     """
#     coord = list(prev_coord)
#     i = len(prev_coord) - 1
#     while coord[i] + delta >= chunk_sizes[i]:
#         coord[i] += delta
#         delta = coord[i] // chunk_sizes[i]
#         coord[i] %= chunk_sizes[i]
#         i -= 1
#     coord[i] += delta
#     return coord


# iquery --afl --format csv --query "
#   sort(
#     project(
#       apply(
#         limit(
#           project(ASSOC_RIVAS_RESULTS, nobs),
#           1000000),
#         d0,   variant_id,
#         d1,   sub_field_id,
#         d0_o, variant_id / 100000 * 100000,
#         d1_o, sub_field_id / 10 * 10),
#       d0, d1, d0_o, d1_o),
#     d0, d1)
#   " > assoc.1m.csv

# ASSOC_RIVAS_RESULTS
#   <pvalue:double COMPRESSION 'zlib',
#    odds_ratio:double COMPRESSION 'zlib',
#    beta:double COMPRESSION 'zlib',
#    se:double COMPRESSION 'zlib',
#    t_stat:double COMPRESSION 'zlib',
#    z_stat:double COMPRESSION 'zlib',
#    nobs:int64 COMPRESSION 'zlib',
#    note:string COMPRESSION 'zlib'>
#   [variant_id=0:*:0:100000; sub_field_id=0:*:0:10]

# iquery --afl --format csv --query "
#   sort(
#     project(
#       apply(
#         limit(
#           filter(
#             UK_BIOBANK.GENOTYPE,
#             variant_id >= 1000000 and variant_id <= 1001000)
#           1000000),
#         d0, variant_id,
#         d1, individual,
#         d0_o, variant_id,
#         d1_o, (individual - 1) / 487409 * 487409 + 1),
#       d0, d1, d0_o, d1_o),
#     d0, d1)
#   " > genotype.1m.csv

# GENOTYPE
#   <probabilities:uint16 NOT NULL COMPRESSION 'zlib'>
#   [variant_id=0:*:0:1; individual=1:487409:0:487409]


# ---
# Read input file
# ---
def read_file(file_name, n_dims, attr_names=(), attr_types=()):
    dim_names = tuple('d{}'.format(i) for i in range(n_dims))
    dim_dtype = dict((k, numpy.int64) for k in dim_names)

    names = tuple(itertools.chain(dim_names, attr_names))
    dtypes = dict(itertools.chain(dim_dtype.items(),
                                  ((name, dtype)
                                   for (name, dtype) in zip(attr_names,
                                                            attr_types))))

    return pandas.read_csv(file_name, names=names, dtype=dtypes)


# ---
# Write output file
# ---
def write_dile(data, file_name):
    data.to_csv(file_name, header=False, index=False)


# ---
# Add pos column
# ---
def compute_delta(data, origins, chunk_sizes):
    n_dims = len(origins)
    n_atts = len(data.columns) - n_dims

    dim_names = tuple('d{}'.format(i) for i in range(n_dims))
    origin_names = tuple('o{}'.format(i) for i in range(n_dims))
    prev_origin_names = tuple('prev_o{}'.format(i) for i in range(n_dims))
    new_dim_names = tuple('new_d{}'.format(i) for i in range(n_dims))

    # Compute chunk origin
    for i in range(n_dims):
        data[origin_names[i]] = ((data[dim_names[i]] - origins[i])
                                 // chunk_sizes[i]
                                 * chunk_sizes[i]
                                 + origins[i])

    # Compute pos
    data['pos'] = coord2pos_all(data,
                                dim_names,
                                origin_names,
                                chunk_sizes)

    # Shift and assign prev_* columns
    data_prev = data.shift(1, fill_value=0)

    # Set previous values (for delta compuation)
    for i in range(n_dims):
        data[prev_origin_names[i]] = data_prev[origin_names[i]]
    data['prev_pos'] = data_prev['pos']

    # Set new chunk markers
    data['new_chunk'] = False
    for (name, prev_name) in zip(origin_names, prev_origin_names):
        data['new_chunk'] |= (data[name] != data[prev_name])

    # if new_chunk then delta = pos
    # else delta = pos - prev_pos
    data['delta'] = data['pos'] - ~data['new_chunk'] * data['prev_pos']

    # Reset first delta due to shift
    data.loc[0, 'delta'] = data.loc[0, 'pos']

    # Verify that no delta value is negative
    if any(data['delta'] < 0):
        raise Exception(
            "Something's off | Negative delta found\n{}".format(
                data.loc[data['delta'] < 0]))

    # Compute new pos
    # if new_chunk pos = delta
    # else pos = prev_pos + delta
    data['new_pos'] = ~data['new_chunk'] * data['prev_pos'] + data['delta']

    # Compute new coords
    data = data.join(
        pos2coord_all(data,
                      'new_pos',
                      origin_names,
                      chunk_sizes,
                      new_dim_names))

    # Verify that new pos and coords match
    for (name, new_name) in zip(dim_names, new_dim_names):
        if not all(data[name] == data[new_name]):
            raise Exception(
                ("Something's off | Computed {} " +
                 "coords don't match\n{}").format(
                     name,
                     data.loc[data[name] != data[new_name]]))
        if not all(data[name] == data[new_name]):
            raise Exception(
                ("Something's off | Computed {} " +
                 "coords don't match\n{}").format(
                     name,
                     data.loc[data[name] != data[new_name]]))

    # Delete new pos and coords
    del data['new_pos']
    for name in new_dim_names:
        del data[name]

    return data


# ---
# Write Arrow table
# ---
def write_arrow(file_name, dataframe, compression='lz4'):
    table = pyarrow.Table.from_pandas(dataframe)
    table = table.replace_schema_metadata()
    stream = pyarrow.output_stream(file_name, compression=compression)
    writer = pyarrow.ipc.RecordBatchStreamWriter(stream, table.schema)
    writer.write_table(table)
    writer.close()
    stream.close()


# ---
# timeit
# ---

# %timeit data.apply(lambda r: coord2pos(r[:2], r[2:4], chunk_size), axis=1)
# %timeit data.apply(lambda r: pos2coord(r[4], r[2:4], chunk_size), axis=1)

# %timeit coord2pos_all(data, ('d0', 'd1'), ('d0_o', 'd1_o'), chunk_size)
# %timeit pos2coord_all(
#     data, 'pos', ('d0_o', 'd1_o'), chunk_size, ('d0', 'd1'))



# BEGIN_COPYRIGHT
#
# Copyright (C) 2020-2023 Paradigm4 Inc.
# All Rights Reserved.
#
# scidbbridge is a plugin for SciDB, an Open Source Array DBMS
# maintained by Paradigm4. See http://www.paradigm4.com/
#
# scidbbridge is free software: you can redistribute it and/or modify
# it under the terms of the AFFERO GNU General Public License as
# published by the Free Software Foundation.
#
# scidbbridge is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY
# KIND, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
# NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See the
# AFFERO GNU General Public License for the complete license terms.
#
# You should have received a copy of the AFFERO GNU General Public
# License along with scidbbridge. If not, see
# <http://www.gnu.org/licenses/agpl-3.0.html>
#
# END_COPYRIGHT

import boto3
import collections
import itertools
import os
import os.path
import numpy
import pandas
import pyarrow
import scidbpy
from typing import List

from .driver import Driver
from .coord import coord2delta, delta2coord

__version__ = '19.11.7'

type_map_pyarrow = dict(
    [(t.__str__(), t) for t in (pyarrow.binary(),
                                pyarrow.bool_(),
                                pyarrow.int16(),
                                pyarrow.int32(),
                                pyarrow.int64(),
                                pyarrow.int8(),
                                pyarrow.string(),
                                pyarrow.uint16(),
                                pyarrow.uint32(),
                                pyarrow.uint64(),
                                pyarrow.uint8())] +
    [('char', pyarrow.string()),
     ('datetime', pyarrow.timestamp('s')),
     ('double', pyarrow.float64()),
     ('float', pyarrow.float32())])


class Array(object):
    """Wrapper for SciDB array stored externally

    Constructor parameters:

    :param string url: URL of the SciDB array. Supported schemas are
      ``s3://`` and ``file://``.

    :param string schema: SciDB array schema for creating a new
      array. Can be specified as ``string`` or ``scidbpy.Schema``
    """
    # 'schema' is a class variable
    schema = None  # type: scidbpy.Schema

    def __init__(self,
                 url,
                 schema=None,
                 format='arrow',
                 compression='lz4',
                 namespace='public',
                 index_split=100000):
        self.url = url

        if schema is None:
            self._metadata = None
            self._schema = None

        else:                   # Create new array
            if type(schema) is scidbpy.Schema:
                self._schema = schema
            else:
                self._schema = scidbpy.Schema.fromstring(schema)

            self._metadata = {
                'attribute':   'ALL',
                'format':      format,
                'version':     '1',
                'schema':      self._schema.__str__(),
                'compression': None if compression == 'none' else compression,
                'index_split': index_split,
                'namespace':   namespace
            }

            Driver.init_array(url)
            Driver.write_metadata(
                url,
                Array.metadata_to_string(self._metadata.copy()))

    def __iter__(self):
        return (i for i in (self.url, ))

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __repr__(self):
        return ('{}(url={!r})').format(type(self).__name__, *self)

    def __str__(self):
        return self.url

    @property
    def metadata(self):
        if self._metadata is None:
            self._metadata = Array.metadata_from_string(
                Driver.read_metadata(self.url))
        return self._metadata

    @property
    def schema(self):
        if self._schema is None:
            self._schema = scidbpy.Schema.fromstring(
                self.metadata['schema'])
        return self._schema

    def delete(self):
        # Delete metadata file first, deleting large arrays could take sometime
        Driver.delete('{}/metadata'.format(self.url))
        Driver.delete_all(self.url)

    def read_index(self):
        # Read index as Arrow Table
        tables = []
        for index_url in Driver.list('{}/index'.format(self.url)):
            tables.append(
                Driver.read_table(index_url,
                                  Driver.index_format,
                                  Driver.index_compression))

        if len(tables):
            table = pyarrow.concat_tables(tables)

            # Convert Arrow Table index to Pandas DataFrame
            index = table.to_pandas(split_blocks=True, self_destruct=True)
            # https://arrow.apache.org/docs/python/pandas.html#reducing-
            # memory-use-i
            del table
            index.sort_values(by=list(index.columns),
                              inplace=True,
                              ignore_index=True)

            return index

        return pandas.DataFrame()

    def build_index(self):
        dims = self.schema.dims
        index = pandas.DataFrame.from_records(
            map(lambda x: Array.url_to_coords(x, dims),
                Driver.list('{}/chunks'.format(self.url))),
            columns=[d.name for d in dims])
        index.sort_values(by=list(index.columns),
                          inplace=True,
                          ignore_index=True)
        return index

    def write_index(self, index, split_size=None):
        # Check for a DataFrame
        if not isinstance(index, pandas.DataFrame):
            raise Exception("Value provided as argument " +
                            "is not a Pandas DataFrame")

        # Check index columns matches array dimentions
        dim_names = [d.name for d in self.schema.dims]
        if len(index.columns) != len(dim_names):
            raise Exception(
                ("Index columns count {} does not match " +
                 "array dimensions count {}").format(len(index.columns),
                                                     len(dim_names)))

        if not (index.columns == dim_names).all():
            raise Exception(
                ("Index columns {} does not match " +
                 "array dimensions {}").format(index.columns, dim_names))

        # Check for coordinates outside chunk boundaries
        for dim in self.schema.dims:
            vals = index[dim.name]
            if any(vals < dim.low_value):
                raise Exception("Index values smaller than " +
                                "lower bound on dimension " + dim.name)
            if dim.high_value != '*' and any(vals > dim.high_value):
                raise Exception("Index values bigger than " +
                                "upper bound on dimension " + dim.name)
            if (dim.chunk_length != '*'
                    and any((vals - dim.low_value) % dim.chunk_length != 0)):
                raise Exception("Index values misaligned " +
                                "with chunk size on dimension " + dim.name)

        # Check for duplicates
        if index.duplicated().any():
            raise Exception("Duplicate entries")

        index.sort_values(by=list(index.columns),
                          inplace=True,
                          ignore_index=True)

        if split_size is None:
            split_size = int(self.metadata['index_split'])

        index_schema = pyarrow.schema(
            [(d.name, pyarrow.int64(), False) for d in self.schema.dims])
        chunk_size = split_size // len(index.columns)

        # Remove existing index
        Driver.delete_all('{}/index'.format(self.url))

        # Write new index
        i = 0
        for offset in range(0, len(index), chunk_size):
            table = pyarrow.Table.from_pandas(
                index.iloc[offset:offset + chunk_size], index_schema)
            Driver.write_table(table,
                               '{}/index/{}'.format(self.url, i),
                               index_schema,
                               Driver.index_format,
                               Driver.index_compression)
            i += 1

    def get_chunk(self, *argv):
        return Chunk(self, *argv)

    def pandas_to_chunks(self, df: pandas.DataFrame) -> List['Chunk']:
        """Returns a list of chunks containing the given DataFrame.

        Args:
            df: pandas.DataFrame contianing data to convert to chunks
        Raises:
            Exception: If columns (attributes or dimensions) do not match,
            wrong schema or duplicate rows
        """
        # Check that columns match array schema
        dims = [d.name for d in self.schema.dims]
        columns = [a.name for a in self.schema.atts] + dims
        if len(df.columns) != len(columns):
            raise Exception(
                ("Argument columns count {} do not match " +
                    "array attributes plus dimensions count {}").format(
                        len(df.columns), len(columns)))

        if sorted(list(df.columns)) != sorted(columns):
            raise Exception(
                ("Argument columns {} does not match " +
                    "array schema {}").format(df.columns, columns))

        # Use schema order
        df = df[columns]

        # Sort by dimensions
        df = df.sort_values(by=dims, ignore_index=True)

        # Check for duplicates
        if df.duplicated(subset=dims).any():
            raise Exception("Duplicate coordinates")

        # Check for coordinates outside array boundaries
        for dim in self.schema.dims:
            max_dim = df[dim.name].max()
            min_dim = df[dim.name].min()

            if dim.high_value != "*":
                if max_dim > dim.high_value:
                    raise Exception(f"Dimension {dim.name} is " +
                                    "above maximum value.")

            if dim.low_value != "*":
                if min_dim < dim.low_value:
                    raise Exception(f"Dimension {dim.name} is " +
                                    "below minimum value.")

        group_dims = []

        for dim in self.schema.dims:
            # TODO: deal with chunk_length of '*'
            df[f"@{dim.name}"] = numpy.int64(numpy.floor(
                (df[dim.name] - dim.low_value) / dim.chunk_length))
            group_dims.append(f"@{dim.name}")

        chunks = []
        grouped_df = df.groupby(group_dims)
        for c in grouped_df.groups:
            df_chunk = grouped_df.get_group(c)
            if not isinstance(c, collections.abc.Iterable):
                # one-dim
                c = (c,)
            # already have real chunk id, but that isn't what get_chunk wants!
            parts = []
            for (coord, dim) in zip(c, self.schema.dims):
                part = coord + dim.low_value
                part = part * dim.chunk_length
                parts.append(part)
            chunk = self.get_chunk(*parts)
            chunk.from_pandas(df_chunk[columns])
            chunks.append(chunk)

        return chunks

    @staticmethod
    def metadata_from_string(input):
        res = dict(ln.split('\t') for ln in input.strip().split('\n'))
        try:
            if res['compression'] == 'none':
                res['compression'] = None
        except KeyError:
            pass
        return res

    @staticmethod
    def metadata_to_string(input):
        if input['compression'] is None:
            input['compression'] = 'none'
        return '\n'.join('{}\t{}'.format(k, v)
                         for (k, v) in input.items()) + '\n'

    @staticmethod
    def coords_to_url_suffix(coords, dims):
        parts = ['c']
        for (coord, dim) in zip(coords, dims):
            if (coord < dim.low_value or
                    dim.high_value != '*' and coord > dim.high_value):
                raise Exception(
                    ('Coordinate value, {}, is outside of dimension range, '
                     '[{}:{}]').format(
                         coord, dim.low_value, dim.high_value))

            part = coord - dim.low_value
            if part % dim.chunk_length != 0:
                raise Exception(
                    ('Coordinate value, {}, is not a multiple of ' +
                     'chunk size, {}').format(
                         coord, dim.chunk_length))
            part = part // dim.chunk_length
            parts.append(part)
        return '_'.join(map(str, parts))

    @staticmethod
    def url_to_coords(url, dims):
        part = url[url.rindex('/') + 1:]
        return tuple(
            map(lambda x: int(x[0]) * x[1].chunk_length + x[1].low_value,
                zip(part.split('_')[1:], dims)))


class Chunk(object):
    """Wrapper for SciDB array chunk stored externally"""

    def __init__(self, array, *argv):
        self.array = array
        self.coords = argv

        if (len(argv) == 1 and
                type(argv[0]) is pandas.core.series.Series):
            argv = tuple(argv[0])

        dims = self.array.schema.dims
        if len(argv) != len(dims):
            raise Exception(
                ('Number of arguments, {}, does not match the number of ' +
                 'dimensions, {}. Please specify one start coordiante for ' +
                 'each dimension.').format(len(argv),
                                           len(self.array.schema.dims)))

        part = Array.coords_to_url_suffix(self.coords, dims)
        self.url = '{}/chunks/{}'.format(self.array.url, part)
        self._table = None

    def __iter__(self):
        return (i for i in (self.array, self.url))

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __repr__(self):
        return ('{}(array={!r}, url={!r})').format(
            type(self).__name__, *self)

    def __str__(self):
        return self.url

    @property
    def table(self):
        if self._table is None:
            self._table = Driver.read_table(
                self.url,
                format=self.array.metadata['format'],
                compression=self.array.metadata['compression'])
        return self._table

    def to_pandas(self):
        return delta2coord(
            self.table.to_pandas(), self.array.schema, self.coords)

    def from_pandas(self, pd):
        # Check for a DataFrame
        if not isinstance(pd, pandas.DataFrame):
            raise Exception("Value provided as argument " +
                            "is not a Pandas DataFrame")

        # Check for empty DataFrame
        if pd.empty:
            raise Exception("Pandas DataFrame is empty. " +
                            "Nothing to do.")

        # Check that columns match array schema
        dims = [d.name for d in self.array.schema.dims]
        columns = [a.name for a in self.array.schema.atts] + dims
        if len(pd.columns) != len(columns):
            raise Exception(
                ("Argument columns count {} do not match " +
                 "array attributes plus dimensions count {}").format(
                     len(pd.columns), len(columns)))

        if sorted(list(pd.columns)) != sorted(columns):
            raise Exception(
                ("Argument columns {} does not match " +
                 "array schema {}").format(pd.columns, columns))

        # Use schema order
        pd = pd[columns]

        # Sort by dimensions
        pd = pd.sort_values(by=dims, ignore_index=True)

        # Check for duplicates
        if pd.duplicated(subset=dims).any():
            raise Exception("Duplicate coordinates")

        # Check for coordinates outside chunk boundaries
        for (coord, dim) in zip(self.coords, self.array.schema.dims):
            vals = pd[dim.name]
            if (vals.iloc[0] < coord or
                    vals.iloc[-1] >= coord + dim.chunk_length):
                raise Exception("Coordinates outside chunk boundaries")

        # Build schema
        schema = pyarrow.schema(
            [(a.name, type_map_pyarrow[a.type_name], not a.not_null)
             for a in self.array.schema.atts] +
            [('@delta', pyarrow.int64(), False)])

        pd['@delta'] = coord2delta(pd, self.array.schema.dims, self.coords)

        self._table = pyarrow.Table.from_pandas(pd, schema)
        self._table = self._table.replace_schema_metadata()

    def save(self):
        Driver.write_table(self._table,
                           self.url,
                           self._table.schema,
                           self.array.metadata['format'],
                           self.array.metadata['compression'])
